import logging


def log_message(msg):
    msg = "Trading API: " + msg
    return msg


def log_error(msg):
    logging.error(log_message(msg))
    return log_message(msg)


def log_exception(e, msg):
    logging.exception(log_message(msg))
    log_message(msg)


def log_warning(msg):
    logging.warning(log_message(msg))
    return log_message(msg)



from ..base import Server

# Server
PROTOCOL = 'https'
HOST = 'api.cryptomkt.com'
VERSION = 'v1'


# CryptoMKT server
class CryptoMKTServer(Server):
    def __init__(self):
        super().__init__(PROTOCOL, HOST, VERSION)



from ..base import Server

# API Server
PROTOCOL = 'https'
HOST = 'api.kraken.com'
VERSION = '0'


# Kraken API server
class KrakenServer(Server):

    def __init__(self):
        super().__init__(PROTOCOL, HOST, VERSION)



# local
from ..base import Client, Server
from ..common import clean_parameters

# API Server
PROTOCOL = 'https'
HOST = 'api.coinmarketcap.com'
VERSION = 'v1'


class CoinMarketCap(Client):

    def __init__(self, timeout: int=120):
        super().__init__(Server(PROTOCOL, HOST, VERSION), timeout)
        self._currencies = None

    def ticker(self,
               currency: str=None,
               convert: str=None,
               start: int=None,
               limit: int=None):
        params = {
            'start': start,
            'limit': limit,
            'convert': convert,
        }
        if currency:
            if len(currency) == 3:
                currency = self._get_symbol(currency)['value']
            url = self.url_for('ticker/%s/', currency)
            data = self.get(url, params=params)[0]
        else:
            url = self.url_for('ticker/')
            data = self.get(url, params=params)
        return data

    def price(self,
              currency: str,
              convert: str=None):
        ticker = self.ticker(currency, convert)
        return float(ticker[f"price_{convert or 'usd'}".lower()])

    def stats(self, convert: str=None):
        params = {
            'convert': convert,
        }
        url = self.url_for('global/')
        data = self.get(url, params=params)
        return data

    def _get_currencies(self):
        ticker = self.ticker()
        return {currency['symbol']: dict(value=currency['id'], decimals=8)
                for currency in ticker}

    def _get_symbol(self, currency: str):
        if self._currencies is None:
            self._currencies = self._get_currencies()
        return self._currencies[currency.upper()]

    def get(self, url: str, headers: dict=None, params: dict=None):
        clean_params = clean_parameters(params)
        return super(CoinMarketCap, self).get(url, params=clean_params)



# local
from ..base import Client, Server
from ..common import format_date_iso, format_datetime_iso

# API Server
PROTOCOL = 'https'
HOST = 'openexchangerates.org/api/'


class OXR(Client):

    error_key = 'error'

    def __init__(self, app_id: str, timeout: int=120):
        super().__init__(Server(PROTOCOL, HOST), timeout)
        self.APP_ID = str(app_id)

    def currencies(self):
        """
        Get a JSON list of all currency symbols available from the Open
        Exchange Rates API, along with their full names.
        ref. https://oxr.readme.io/docs/currencies-json
        """
        url = self.url_for('currencies.json')
        data = self.get(url)
        return data

    def latest(self, 
               base: str=None, 
               symbols: list=None):
        """
        Get latest data.
        ref. https://oxr.readme.io/docs/latest-json
        """
        url = self.url_for('latest.json')
        data = self._get_exchange_rates(url, base, symbols)
        return data

    def historical(self,
                   date_for: str,
                   base: str=None, 
                   symbols: list=None):
        """
        Get daily historical data
        ref. https://oxr.readme.io/docs/historical-json
        """
        date_for = format_date_iso(date_for)
        url = self.url_for('historical/%s.json', date_for)
        data = self._get_exchange_rates(url, base, symbols)
        return data

    def time_series(self,
                    start: str,
                    end: str,
                    base: str=None, 
                    symbols: list=None):
        """
        Get time-series data.
        ref. https://oxr.readme.io/docs/time-series-json
        """
        params = {
            'start': format_date_iso(start),
            'end': format_date_iso(end),
        }
        url = self.url_for('time-series.json')
        data = self._get_exchange_rates(url, base, symbols, params)
        return data

    def convert(self,
                value: int,
                from_symbol: str,
                to_symbol: str):
        """
        Convert any money value from one currency to another at the latest
        API rates.
        ref. https://oxr.readme.io/docs/convert
        """
        url = self.url_for('convert/%s/%s/%s', (value, from_symbol, to_symbol))
        data = self._get_exchange_rates(url)
        return data

    def ohlc(self,
             start_time: str,
             period: str,
             symbols: list=None,
             base: str=None):
        """
        Get historical open, high, low, close (OHLC) and average data.
        ref. https://oxr.readme.io/docs/ohlc-json
        """
        params = {
            'start_time': format_datetime_iso(start_time),
            'period': period,
        }
        url = self.url_for('ohlc.json')
        data = self._get_exchange_rates(url, base, symbols, params)
        return data

    def _get_exchange_rates(self,
                            url: str,
                            base: str=None,
                            symbols: list=None,
                            params: dict=None):
        if params is None:
            params = dict()
        params['app_id'] = self.APP_ID
        if base is not None:
            params['base'] = base
        if isinstance(symbols, list) or isinstance(symbols, tuple):
            symbols = ','.join(symbols)
        if symbols is not None:
            params['symbols'] = symbols
        return self.get(url, params=params)



import json
from enum import Enum
from urllib.parse import urlparse

# pip
import requests

# local
from . import errors


class Server(object):

    def __init__(self, protocol, host, version=None):
        url = f'{protocol}://{host}'
        if version:
            url = f'{url}/{version}'

        self.PROTOCOL = protocol
        self.HOST = host
        self.VERSION = version
        self.URL = url


class Client(object):

    error_key = ''

    def __init__(self, server: Server, timeout=30):
        self.SERVER = server
        self.TIMEOUT = timeout

    def get(self, url, headers=None, params=None):
        response = self._request('get', url, headers=headers, params=params)
        return response

    def put(self, url, headers, data):
        response = self._request('put', url, headers=headers, data=data)
        return response

    def post(self, url, headers, data):
        response = self._request('post', url, headers=headers, data=data)
        return response

    def _request(self, method, url, headers, params=None, data=None):
        data = self._encode_data(data)
        response = requests.request(
            method,
            url,
            headers=headers,
            params=params,
            data=data,
            verify=True,
            timeout=self.TIMEOUT)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise errors.InvalidResponse(response) from e
        json_resp = self._resp_to_json(response)
        return json_resp

    def _encode_data(self, data):
        data = json.dumps(data) if data else data
        return data

    def _resp_to_json(self, response):
        try:
            json_resp = response.json()
        except json.decoder.JSONDecodeError as e:
            raise errors.DecodeError() from e
        if isinstance(json_resp, dict):
            if bool(json_resp.get(self.error_key, False)):
                raise errors.InvalidResponse(response)
        return json_resp

    def url_for(self, path, path_arg=None):
        url = f'{self.SERVER.URL}/{path}'
        if path_arg:
            url = url % path_arg
        return url

    def url_path_for(self, path, path_arg=None):
        url = self.url_for(path, path_arg)
        path = urlparse(url).path
        return url, path


class _Enum(Enum):
    @staticmethod
    def _format_value(value):
        return str(value).upper()

    @classmethod
    def check(cls, value):
        if value is None:
            return value
        if type(value) is cls:
            return value
        try:
            return cls[cls._format_value(value)]
        except KeyError:
            return cls._missing_(value)

    def __str__(self):
        return self.value


class _Currency(_Enum):
    @property
    def value(self):
        return super(_Currency, self).value['value']

    @property
    def decimals(self):
        return super(_Currency, self).value.get('decimals', 2)


class _Market(_Enum):
    @staticmethod
    def _format_value(value):
        value = str(value).replace('-', '')
        value = f'{value[:3]}_{value[3:]}'
        return value.upper()

    @property
    def value(self):
        return super(_Market, self).value['value']

    @property
    def base(self):
        return super(_Market, self).value['base']

    @property
    def quote(self):
        return super(_Market, self).value['quote']



# local
from .server import KrakenServer
from ..base import Client


class KrakenPublic(Client):

    error_key = 'error'

    def __init__(self, timeout: int=30):
        super().__init__(KrakenServer(), timeout)

    def server_time(self):
        url = self.url_for('public/Time')
        return self.get(url)

    def assets(self):
        url = self.url_for('public/Assets')
        return self.get(url)

    def asset_pairs(self):
        url = self.url_for('public/AssetPairs')
        return self.get(url)

    def ticker(self, symbol: str):
        params = {
            'pair': str(symbol),
        }
        url = self.url_for('public/Ticker')
        return self.get(url, params=params)

    def ohcl(self,
             symbol: str,
             interval: int=None,
             since: str=None):
        params = {
            'pair': str(symbol),
            'interval': interval,
            'since': since,
        }
        url = self.url_for('public/OHCL')
        return self.get(url, params=params)

    def order_book(self,
                   symbol: str,
                   count: int=None):
        params = {
            'pair': str(symbol),
            'count': count,
        }
        url = self.url_for('public/Depth')
        return self.get(url, params=params)

    def trades(self,
               symbol: str,
               since: str=None):
        params = {
            'pair': str(symbol),
            'since': since,
        }
        url = self.url_for('public/Trades')
        return self.get(url, params=params)

    def spread(self,
               symbol: str,
               since: str=None):
        params = {
            'pair': str(symbol),
            'since': since,
        }
        url = self.url_for('public/Spread')
        return self.get(url, params=params)



# local
from .server import BitstampServer
from ..base import Client


class BitstampPublic(Client):

    error_key = 'error'

    def __init__(self, timeout=30):
        super().__init__(BitstampServer(), timeout)

    def url_for(self, path, path_arg=None, version=2):
        if version == 1:
            url = f'{self.SERVER.URL}/{path}/'
        else:
            url = f'{self.SERVER.URL}/v{version}/{path}/'
        if path_arg:
            url = url % path_arg
        return url

    def ticker(self, currency_pair: str):
        """
        Returns ticker dictionary.
        """
        url = self.url_for('ticker/%s', currency_pair)
        return self.get(url)

    def ticker_hour(self, currency_pair: str):
        """
        Returns dictionary of the average ticker of the past hour.
        """
        url = self.url_for('ticker_hour/%s', currency_pair)
        return self.get(url)

    def order_book(self, currency_pair: str):
        """
        Returns dictionary with "bids" and "asks".

        Each is a list of open orders and each order is represented as a list
        of price and amount.
        """
        url = self.url_for('order_book/%s', currency_pair)
        return self.get(url)

    def transactions(self, currency_pair: str, time_interval: str=None):
        """
        Returns transactions for the last 'timedelta' seconds.

        Parameter time is specified by one of two values of TransRange class.
        """
        params = {'time': str(time_interval)} if time_interval else None
        url = self.url_for('transactions/%s', currency_pair)
        return self.get(url, params=params)

    def trading_pairs_info(self):
        """
        Returns list of dictionaries specifying details of each available
        trading pair:

            {
                'description':'Litecoin / U.S. dollar',
                'name':'LTC/USD',
                'url_symbol':'ltcusd',
                'trading':'Enabled',
                'minimum_order':'5.0 USD',
                'counter_decimals':2,
                'base_decimals':8
            },
        """
        url = self.url_for('trading-pairs-info')
        return self.get(url)

    def conversion_rate_usd_eur(self):
        """
        Returns simple dictionary:

            {'buy': 'buy conversion rate', 'sell': 'sell conversion rate'}
        """
        url = self.url_for('eur_usd', version=1)
        return self.get(url)



import base64
import hashlib
import hmac
import json
from datetime import datetime

# local
from . import constants as _c
from . import models as _m
from ..common import build_route, check_keys, gen_nonce
from .client_public import BudaPublic


class BudaAuth(BudaPublic):

    def __init__(self, key: str=False, secret: str=False,
                 timeout: int=30, host: str=None, return_json: bool=False):
        super().__init__(timeout, host, return_json)
        check_keys(key, secret)
        self.KEY = str(key)
        self.SECRET = str(secret)

    def quotation(self,
                  market_id: str,
                  quotation_type: str,
                  amount: float,
                  limit: float=None):
        payload = {
            'quotation': {
                'type': str(quotation_type),
                'amount': str(amount),
                'limit': str(limit) if limit else None,
            },
        }
        url, path = self.url_path_for('markets/%s/quotations', market_id)
        data = self._sign_and_post(url, path, payload=payload)
        if self.return_json:
            return data
        return _m.Quotation.create_from_json(data['quotation'])

    def quotation_market(self,
                         market_id: str,
                         quotation_type: str,
                         amount: float):
        return self.quotation(
            market_id=market_id, quotation_type=quotation_type,
            amount=amount, limit=None)

    def quotation_limit(self,
                        market_id: str,
                        quotation_type: str,
                        amount: float,
                        limit: float):
        return self.quotation(
            market_id=market_id, quotation_type=quotation_type,
            amount=amount, limit=limit)

    # REPORTS -----------------------------------------------------------------
    def _report(self,
                market_id: str,
                report_type: _c.ReportType,
                start_at: datetime=None,
                end_at: datetime=None):
        if isinstance(start_at, datetime):
            start_at = int(start_at.timestamp())
        if isinstance(end_at, datetime):
            end_at = int(end_at.timestamp())
        params = {
            'report_type': str(report_type),
            'from': start_at,
            'to': end_at,
        }
        url, path = self.url_path_for('markets/%s/reports', market_id)
        data = self._sign_and_get(url, path, params=params)
        return data

    def report_average_prices(self,
                              market_id: str,
                              start_at: datetime=None,
                              end_at: datetime=None):
        data = self._report(
            market_id=market_id, report_type=_c.ReportType.AVERAGE_PRICES,
            start_at=start_at, end_at=end_at)
        if self.return_json:
            return data
        return [_m.AveragePrice.create_from_json(report)
                for report in data['reports']]

    def report_candlestick(self,
                           market_id: str,
                           start_at: datetime = None,
                           end_at: datetime = None):
        data = self._report(
            market_id=market_id, report_type=_c.ReportType.CANDLESTICK,
            start_at=start_at, end_at=end_at)
        if self.return_json:
            return data
        return [_m.Candlestick.create_from_json(report)
                for report in data['reports']]

    # BALANCES-----------------------------------------------------------------
    def balance(self, currency: str):
        url, path = self.url_path_for('balances/%s', currency)
        data = self._sign_and_get(url, path)
        if self.return_json:
            return data
        return _m.Balance.create_from_json(data['balance'])

    def balance_event_pages(self,
                            currencies: list,
                            event_names: list,
                            page: int=None,
                            per_page: int=None,
                            relevant: bool=None):
        currencies = [str(c) for c in currencies]
        event_names = [str(e) for e in event_names]
        params = {
            'currencies[]': currencies,
            'event_names[]': event_names,
            'page': page,
            'per': per_page,
            'relevant': relevant,
        }
        url, path = self.url_path_for('balance_events')
        data = self._sign_and_get(url, path, params=params)
        if self.return_json:
            return data
        # TODO: Response only contains a 'total_count' field instead of meta
        return _m.BalanceEventPages.create_from_json(
            data['balance_events'], data['total_count'], page)

    # ORDERS ------------------------------------------------------------------
    def new_order(self,
                  market_id: str,
                  order_type: str,
                  price_type: str,
                  amount: float,
                  limit: float=None):
        payload = {
            'type': str(order_type),
            'price_type': str(price_type),
            'amount': str(amount),
            'limit': str(limit) if limit else None,
        }
        return self.new_order_payload(market_id, payload)

    def new_order_payload(self, market_id: str, payload):
        url, path = self.url_path_for('markets/%s/orders', market_id)
        data = self._sign_and_post(url, path, payload=payload)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['order'])

    def order_pages(self,
                    market_id: str,
                    page: int=None,
                    per_page: int=None,
                    state: str=None,
                    minimum_exchanged: float=None):
        if per_page and per_page > _c.ORDERS_LIMIT:
            msg = f"Param 'per_page' must be <= {_c.ORDERS_LIMIT}"
            raise ValueError(msg)
        params = {
            'per': per_page,
            'page': page,
            'state': str(state) if state else None,
            'minimum_exchanged': minimum_exchanged,
        }
        url, path = self.url_path_for('markets/%s/orders', market_id)
        data = self._sign_and_get(url, path, params=params)
        if self.return_json:
            return data
        return _m.OrderPages.create_from_json(data['orders'], data.get('meta'))

    def order_details(self, order_id: int):
        url, path = self.url_path_for('orders/%s', order_id)
        data = self._sign_and_get(url, path)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['order'])

    def cancel_order(self, order_id: int):
        payload = {
            'state': _c.OrderState.CANCELING.value,
        }
        url, path = self.url_path_for('orders/%s', order_id)
        data = self._sign_and_put(url, path, payload=payload)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['order'])

    # PAYMENTS ----------------------------------------------------------------
    def _transfers(self, currency: str, path, model, key):
        url, path = self.url_path_for(path, currency)
        data = self._sign_and_get(url, path)
        if self.return_json:
            return data
        return [model.create_from_json(transfer)
                for transfer in data[key]]

    def withdrawals(self, currency: str):
        return self._transfers(currency=currency,
                               path='currencies/%s/withdrawals',
                               model=_m.Withdrawal, key='withdrawals')

    def deposits(self, currency: str):
        return self._transfers(currency=currency,
                               path='currencies/%s/deposits',
                               model=_m.Deposit, key='deposits')

    # TODO: UNTESTED
    def withdrawal(self,
                   currency: str,
                   amount: float,
                   target_address: str=None,
                   amount_includes_fee: bool=True,
                   simulate: bool=False):
        payload = {
            'withdrawal_data': {
                'target_address': target_address,
            },
            'amount': str(amount),
            'currency': str(currency),
            'simulate': simulate,
            'amount_includes_fee': amount_includes_fee,
        }
        url, path = self.url_path_for('currencies/%s/withdrawals', currency)
        data = self._sign_and_post(url, path, payload=payload)
        if self.return_json:
            return data
        return _m.Withdrawal.create_from_json(data['withdrawal'])

    def simulate_withdrawal(self,
                            currency: str,
                            amount: float,
                            amount_includes_fee: bool=True):
        return self.withdrawal(
            currency=currency, amount=amount, target_address=None,
            amount_includes_fee=amount_includes_fee, simulate=True)

    # PRIVATE METHODS ---------------------------------------------------------
    def _sign_payload(self, method, path, params=None, payload=None):

        route = build_route(path, params)
        nonce = gen_nonce()

        if payload:
            j = json.dumps(payload).encode('utf-8')
            encoded_body = base64.standard_b64encode(j).decode('utf-8')
            msg = ' '.join([method, route, encoded_body, nonce])
        else:
            msg = ' '.join([method, route, nonce])

        h = hmac.new(key=self.SECRET.encode('utf-8'),
                     msg=msg.encode('utf-8'),
                     digestmod=hashlib.sha384)

        signature = h.hexdigest()

        return {
            'X-SBTC-APIKEY': self.KEY,
            'X-SBTC-NONCE': nonce,
            'X-SBTC-SIGNATURE': signature,
            'Content-Type': 'application/json',
        }

    def _sign_and_post(self, url, path, params=None, payload=None):
        signed_payload = self._sign_payload('POST', path, params, payload)
        return self.post(url, headers=signed_payload, data=payload)

    def _sign_and_put(self, url, path, params=None, payload=None):
        signed_payload = self._sign_payload('PUT', path, params, payload)
        return self.put(url, headers=signed_payload, data=payload)

    def _sign_and_get(self, url, path, params=None, payload=None):
        signed_payload = self._sign_payload('GET', path, params, payload)
        return self.get(url, headers=signed_payload, params=params)



# local
from . import constants as _c
from . import models as _m
from ..base import Client
from .server import BudaServer


class BudaPublic(Client):

    error_key = 'message'

    def __init__(self, timeout: int=30, host: str=None,
                 return_json: bool=False):
        super().__init__(BudaServer(host), timeout)
        self.return_json = return_json

    def markets(self):
        url, path = self.url_path_for('markets')
        data = self.get(url)
        if self.return_json:
            return data
        return [_m.Market.create_from_json(market)
                for market in data['markets']]

    def market_details(self, market_id: str):
        url = self.url_for('markets/%s', market_id)
        data = self.get(url)
        if self.return_json:
            return data
        return _m.Market.create_from_json(data['market'])

    def ticker(self, market_id: str):
        url = self.url_for('markets/%s/ticker', market_id)
        data = self.get(url)
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data['ticker'])

    def order_book(self, market_id: str):
        url = self.url_for('markets/%s/order_book', market_id)
        data = self.get(url)
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(data['order_book'])

    def trades(self,
               market_id: str,
               timestamp: int=None):
        params = {
            'timestamp': timestamp,
        }
        url, path = self.url_path_for('markets/%s/trades', market_id)
        data = self.get(url, params=params)
        if self.return_json:
            return data
        return _m.Trades.create_from_json(data['trades'])



import hashlib
import hmac
import time

# local
from . import constants as _c
from . import models as _m
from .client_public import CryptoMKTPublic
from ..common import check_keys, clean_parameters


class CryptoMKTAuth(CryptoMKTPublic):

    def __init__(self, key: str=False, secret: str=False, timeout: int=30,
                 return_json=False):
        super().__init__(timeout, return_json)
        check_keys(key, secret)
        self.KEY = str(key)
        self.SECRET = str(secret)

    # BALANCE------------------------------------------------------------------
    def balance(self):
        url, path = self.url_path_for('balance')
        headers = self._sign_payload(path=path)
        data = self.get(url, headers=headers)
        if self.return_json:
            return data
        return _m.Balance.create_from_json(data['data'])

    def wallet_balance(self, currency: str):
        balance = self.balance()
        return getattr(balance, str(currency))

    # ORDERS-------------------------------------------------------------------
    def active_orders(self,
                      market_id: str,
                      page: int=None,
                      limit: int=_c.ORDERS_LIMIT):
        params = {
            'market': str(market_id),
            'page': page,
            'limit': limit,
        }
        url, path = self.url_path_for('orders/active')
        headers = self._sign_payload(path=path)
        data = self.get(url, headers=headers, params=params)
        if self.return_json:
            return data
        return _m.Orders.create_from_json(data['data'], data['pagination'])

    def executed_orders(self,
                        market_id: str,
                        page: int=None,
                        limit: int=_c.ORDERS_LIMIT):
        params = {
            'market': str(market_id),
            'page': page,
            'limit': limit,
        }
        url, path = self.url_path_for('orders/executed')
        headers = self._sign_payload(path=path)
        data = self.get(url, headers=headers, params=params)
        if self.return_json:
            return data
        return _m.Orders.create_from_json(data['data'], data['pagination'])

    def create_order(self,
                     market_id: str,
                     order_type: str,
                     amount: float,
                     price: float):
        payload = {
            'market': str(market_id),
            'type': str(order_type),
            'amount': amount,
            'price': price,
        }
        url, path = self.url_path_for('orders/create')
        headers = self._sign_payload(path=path, payload=payload)
        data = self.post(url, headers=headers, data=payload)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['data'])

    def order_status(self, order_id: str):
        params = {
            'id': order_id,
        }
        url, path = self.url_path_for('orders/status')
        headers = self._sign_payload(path=path)
        data = self.get(url, headers=headers, params=params)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['data'])

    def cancel_order(self, order_id: str):
        payload = {
            'id': order_id,
        }
        url, path = self.url_path_for('orders/cancel')
        headers = self._sign_payload(path=path, payload=payload)
        data = self.post(url, headers=headers, data=payload)
        if self.return_json:
            return data
        return _m.Order.create_from_json(data['data'])

    # PAYMENTS-----------------------------------------------------------------
    # TODO: Not tested
    def create_payment(self,
                       amount: float,
                       currency: str,
                       account_email: str,
                       external_id: str=None,
                       callback_url: str=None,
                       error_url: str=None,
                       success_url: str=None,
                       refund_email: str=None):
        payload = {
            'to_receive': amount,
            'to_receive_currency': str(currency),
            'payment_receiver': account_email,
            'external_id': external_id,
            'callback_url': callback_url,
            'error_url': error_url,
            'success_url': success_url,
            'refund_email': refund_email,
        }
        payload = clean_parameters(payload)
        url, path = self.url_path_for('payment/new_order')
        headers = self._sign_payload(path=path, payload=payload)
        data = self.post(url, headers=headers, data=payload)
        if self.return_json:
            return data
        return data

    # TODO: Not tested
    def payment_status(self, payment_id: str):
        params = {
            'id': payment_id,
        }
        url, path = self.url_path_for('payment/status')
        headers = self._sign_payload(path=path)
        data = self.get(url, headers=headers, params=params)
        return data

    # PRIVATE METHODS ---------------------------------------------------------
    def _sign_payload(self, path: str, payload: dict=None):

        timestamp = str(int(time.time()))
        msg = timestamp + path

        if payload:
            for value in [str(payload[k]) for k in sorted(payload.keys())]:
                msg += value

        signature = hmac.new(key=self.SECRET.encode('utf-8'),
                             msg=msg.encode('utf-8'),
                             digestmod=hashlib.sha384).hexdigest()

        # Request fails with 'get_requests_not_allowed' when
        # 'Content-Type': 'application/json' is present
        return {
            'X-MKT-APIKEY': self.KEY,
            'X-MKT-SIGNATURE': signature,
            'X-MKT-TIMESTAMP': timestamp,
        }

    # Request fails with 'get_requests_not_allowed' when
    # json.dumps()' is used
    def _encode_data(self, data):
        return data



from . import logger


class Error(Exception):
    pass


class Unavailable(Error):
    pass


class TradingAPIError(Error):
    def __init__(self, message):
        self.message = message
        logger.log_error(self.message)


class InvalidResponse(TradingAPIError):
    def __init__(self, response):
        msg = ('InvalidResponse (Code: {r.status_code}): '
               'Message: {r.text} ({r.url})'.
               format(r=response))
        super(InvalidResponse, self).__init__(msg)


class DecodeError(TradingAPIError):
    def __init__(self):
        msg = 'DecodeError: Unable to decode JSON from response (no content).'
        super(DecodeError, self).__init__(msg)



from ..base import Server

# API Server
PROTOCOL = 'https'
HOST = 'www.buda.com/api'
VERSION = 'v2'


# Buda API server
class BudaServer(Server):

    def __init__(self, host):
        host = host or HOST
        super().__init__(PROTOCOL, host, VERSION)



import time
from datetime import date, datetime, timedelta
from urllib.parse import urlencode

# local
from . import logger


def gen_nonce():
    # Sleeps 200ms to avoid flooding the server with requests.
    time.sleep(0.2)
    # Get a str from the current time in microseconds.
    return str(int(time.time() * 1E6))


def check_keys(key, secret):
    if not key or not secret:
        msg = 'API Key and Secret are needed!'
        logger.log_error(msg)
        raise ValueError(msg)


def build_parameters(parameters):
    if parameters:
        p = clean_parameters(parameters)
        return urlencode(p, True)
    else:
        return None


def build_route(path, params=None):
    built_params = build_parameters(params)
    if built_params:
        return f'{path}?{built_params}'
    else:
        return path


def clean_parameters(parameters: dict):
    if parameters:
        return {k: v for k, v in parameters.items() if v is not None}


def date_range(start_date, end_date):
    for n in range(int((end_date - start_date).days)):
        yield start_date + timedelta(n)


def current_utc_date():
    return datetime.utcnow().date()


def format_date_iso(date_value):
    if date is None:
        return None
    if isinstance(date_value, datetime):
        date_value = date_value.date()
    if isinstance(date_value, date):
        date_value = date_value.isoformat()
    return date_value


def format_datetime_iso(date_value):
    if date is None:
        return None
    if isinstance(date_value, date):
        date_value = datetime(
            date_value.year, date_value.month, date_value.day)
    if isinstance(date_value, datetime):
        date_value = date_value.isoformat()
    return date_value



from ..base import _Enum, _Market, _Currency

# Limits
ORDERS_LIMIT = 100


class Currency(_Currency):
    ARS = dict(value='ARS')
    BRL = dict(value='BRL')
    CLP = dict(value='CLP')
    ETH = dict(value='ETH', decimals=18)
    EUR = dict(value='EUR')


class OrderType(_Enum):
    BUY = 'buy'
    SELL = 'sell'


class Market(_Market):
    ETH_ARS = dict(value='ETHARS', base=Currency.ETH, quote=Currency.ARS)
    ETH_CLP = dict(value='ETHCLP', base=Currency.ETH, quote=Currency.CLP)
    ETH_EUR = dict(value='ETHEUR', base=Currency.ETH, quote=Currency.EUR)


class TimeFrame(_Enum):
    DAYS_1 = 1440
    HOURS_1 = 60
    MINUTES_15 = 15
    MINUTES_5 = 5
    MINUTES_1 = 1

    @staticmethod
    def _format_value(value):
        values = {
            1440: 'DAYS_1',
            60: 'HOURS_1',
            15: 'MINUTES_15',
            5: 'MINUTES_5',
            1: 'MINUTES_1',
        }
        return values[value]



from ..base import Server

# API Server
PROTOCOL = 'https'
HOST = 'api.bitfinex.com'


# Bitfinex API server
class BitfinexServer(Server):

    def __init__(self, version: int):
        super().__init__(PROTOCOL, HOST, version=f'v{version}')



import base64
import hashlib
import hmac
from urllib.parse import urlencode

# local
from .client_public import KrakenPublic
from ..common import check_keys, clean_parameters, gen_nonce


class KrakenAuth(KrakenPublic):

    def __init__(self, key: str=False, secret: str=False, timeout: int=30):
        super().__init__(timeout)
        check_keys(key, secret)
        self.KEY = str(key)
        self.SECRET = str(secret)

    # Private user data -------------------------------------------------------
    # Get account balance.
    def balance(self):
        url, path = self.url_path_for('private/Balance')
        return self._sign_and_post(url, path, payload={})

    # Get trade balance.
    def trade_balance(self,
                      asset: str=None,
                      asset_class: str=None):
        payload = {
            'asset': str(asset) if asset else None,
            'aclass': str(asset_class) if asset_class else None,
        }
        url, path = self.url_path_for('private/TradeBalance')
        return self._sign_and_post(url, path, payload)

    # Get open orders.
    def open_orders(self,
                    include_trades: bool=None,
                    userref: str=None):
        payload = {
            'trades': include_trades,
            'userref': userref,
        }
        url, path = self.url_path_for('private/OpenOrders')
        return self._sign_and_post(url, path, payload)

    # Get closed orders.
    def closed_orders(self,
                      include_trades: bool=None,
                      userref: str=None,
                      start: int=None,
                      end: int=None,
                      ofs: int=None,
                      closetime: int=None):
        payload = {
            'trades': include_trades,
            'userref': userref,
            'start': start,
            'end': end,
            'ofs': ofs,
            'closetime': closetime,
        }
        url, path = self.url_path_for('private/ClosedOrders')
        return self._sign_and_post(url, path, payload)

    # Query orders info.
    def query_orders(self,
                     txid: list,
                     include_trades: bool=None,
                     userref: str=None):
        payload = {
            'txid': txid,
            'trades': include_trades,
            'userref': userref,
        }
        url, path = self.url_path_for('private/QueryOrders')
        return self._sign_and_post(url, path, payload)

    # Get trades history.
    def trades_history(self,
                       trade_type: str=None,
                       include_trades: bool=None,
                       start: int=None,
                       end: int=None,
                       ofs: int=None):
        payload = {
            'type': str(trade_type) if trade_type else None,
            'trades': include_trades,
            'start': start,
            'end': end,
            'ofs': ofs,
        }
        url, path = self.url_path_for('private/TradesHistory')
        return self._sign_and_post(url, path, payload)

    # Query trades info.
    def query_trades(self,
                     txid: list,
                     include_trades: bool=None):
        payload = {
            'txid': txid,
            'trades': include_trades,
        }
        url, path = self.url_path_for('private/QueryTrades')
        return self._sign_and_post(url, path, payload)

    # Query trades info.
    def open_positions(self,
                       txid: list,
                       include_pl: bool=None):
        payload = {
            'txid': txid,
            'docalcs': include_pl,
        }
        url, path = self.url_path_for('private/OpenPositions')
        return self._sign_and_post(url, path, payload)

    # Get ledgers info.
    def ledgers(self,
                asset_class: str=None,
                asset: str=None,
                ledger_type: str=None,
                start: int=None,
                end: int=None,
                ofs: int=None):
        payload = {
            'aclass': str(asset_class) if asset_class else None,
            'asset': str(asset) if asset else None,
            'type': str(ledger_type) if ledger_type else None,
            'start': start,
            'end': end,
            'ofs': ofs,
        }
        url, path = self.url_path_for('private/Ledgers')
        return self._sign_and_post(url, path, payload)

    # Query ledgers.
    def query_ledgers(self, ledger_id: str):
        payload = {
            'id': ledger_id,
        }
        url, path = self.url_path_for('private/QueryLedgers')
        return self._sign_and_post(url, path, payload)

    # Get trade volume.
    def trade_volume(self,
                     pair: str=None,
                     fee_info: bool=None):
        payload = {
            'pair': str(pair) if pair else None,
            'fee-info': fee_info,
        }
        url, path = self.url_path_for('private/TradeVolume')
        return self._sign_and_post(url, path, payload)

    # Private user trading  ---------------------------------------------------
    # Add standard order
    def add_order(self,
                  pair: str,
                  direction: str,
                  order_type: str,
                  volume: float,
                  price: float=None,
                  price2: float=None,
                  leverage: float=None,
                  oflags: list=None,
                  starttm: int=None,
                  expiretm: int=None,
                  userref: str=None,
                  validate: bool=None,
                  c_ordertype: str=None,
                  c_price: float=None,
                  c_price2: float=None):
        payload = {
            'pair': str(pair),
            'type': str(direction),
            'ordertype': str(order_type),
            'price': price,
            'price2': price2,
            'volume': volume,
            'leverage': leverage,
            'oflags': oflags,
            'starttm': starttm,
            'expiretm': expiretm,
            'userref': userref,
            'validate': validate,
            'close[ordertype]': str(c_ordertype) if order_type else None,
            'close[price]': c_price,
            'close[price2]': c_price2,
        }
        url, path = self.url_path_for('private/AddOrder')
        return self._sign_and_post(url, path, payload)

    # Cancel open order
    def cancel_order(self, txid: str):
        payload = {
            'txid': txid,
        }
        url, path = self.url_path_for('private/CancelOrder')
        return self._sign_and_post(url, path, payload)

    # Private user funding  ---------------------------------------------------
    # Get deposit methods.
    def deposit_methods(self,
                        asset: str,
                        asset_class: str=None):
        payload = {
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
        }
        url, path = self.url_path_for('private/DepositMethods')
        return self._sign_and_post(url, path, payload)

    # Get deposit addresses
    def deposit_addresses(self,
                          asset: str,
                          method: str,
                          asset_class: str=None,
                          new: bool=None):
        payload = {
            'asset': str(asset),
            'method': str(method),
            'aclass': str(asset_class) if asset_class else None,
            'new': new,
        }
        url, path = self.url_path_for('private/DepositAddresses')
        return self._sign_and_post(url, path, payload)

    # Get status of recent deposits
    def deposit_status(self,
                       asset: str,
                       method: str,
                       asset_class: str=None):
        payload = {
            'asset': str(asset),
            'method': str(method),
            'aclass': str(asset_class) if asset_class else None,
        }
        url, path = self.url_path_for('private/DepositStatus')
        return self._sign_and_post(url, path, payload)

    # Get withdrawal information
    def withdraw_info(self,
                      asset: str,
                      amount: float,
                      key: str,
                      asset_class: str=None):
        payload = {
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
            'amount': amount,
            'key': key,
        }
        url, path = self.url_path_for('private/WithdrawInfo')
        return self._sign_and_post(url, path, payload)

    # Withdraw funds
    def withdraw(self,
                 asset: str,
                 amount: float,
                 key: str,
                 asset_class: str=None):
        payload = {
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
            'amount': amount,
            'key': key,
        }
        url, path = self.url_path_for('private/Withdraw')
        return self._sign_and_post(url, path, payload)

    # Get status of recent withdrawals
    def withdraw_status(self,
                        asset: str,
                        method: str,
                        asset_class: str=None):
        payload = {
            'asset': str(asset),
            'method': str(method),
            'aclass': str(asset_class) if asset_class else None,
        }
        url, path = self.url_path_for('private/WithdrawStatus')
        return self._sign_and_post(url, path, payload)

    # Request withdrawal cancellation
    def withdraw_cancel(self,
                        asset: str,
                        refid: str,
                        asset_class: str=None):
        payload = {
            'asset': str(asset),
            'aclass': str(asset_class) if asset_class else None,
            'refid': refid,
        }
        url, path = self.url_path_for('private/WithdrawCancel')
        return self._sign_and_post(url, path, payload)

    # PRIVATE METHODS ---------------------------------------------------------
    def _sign_payload(self, path: str, data: dict=None):
        data['nonce'] = gen_nonce()
        encoded_data = self._encode_data(data)

        # Unicode-objects must be encoded before hashing
        encoded = (str(data['nonce']) + encoded_data).encode()
        message = path.encode() + hashlib.sha256(encoded).digest()

        signature = hmac.new(key=base64.b64decode(self.SECRET),
                             msg=message,
                             digestmod=hashlib.sha512)
        sig_digest = base64.b64encode(signature.digest())

        return {
            'API-Key': self.KEY,
            'API-Sign': sig_digest.decode(),
        }

    def _encode_data(self, data):
        encoded_data = urlencode(data) if data else data
        return encoded_data

    # Packs and sign the payload and send the request with POST.
    def _sign_and_post(self, url: str, path: str, payload: dict):
        payload = clean_parameters(payload)
        signed_payload = self._sign_payload(path, payload)
        return self.post(url, headers=signed_payload, data=payload)



from ..base import Server

# API Server
PROTOCOL = 'https'
HOST = 'www.bitstamp.net/api'


# Bitstamp API server
class BitstampServer(Server):

    def __init__(self):
        super().__init__(PROTOCOL, HOST)



# local
from ..base import Client, Server

# API Server
PROTOCOL = 'https'
HOST = 'bitcoinity.org'


class Bitcoinity(Client):
    def __init__(self, timeout=15):
        super().__init__(Server(PROTOCOL, HOST), timeout)

    def ticker(self,
               currency: str,
               exchange: str,
               span: str):
        params = {
            'currency': currency,
            'exchange': exchange,
            'span': span,
        }
        url = self.url_for('markets/get_ticker')
        data = self.get(url, params=params)
        return data



from datetime import datetime

# local
from . import constants as _c
from . import models as _m
from ..base import Client
from .server import CryptoMKTServer


class CryptoMKTPublic(Client):

    error_key = 'message'

    def __init__(self, timeout: int=30, return_json=False):
        super().__init__(CryptoMKTServer(), timeout)
        self.return_json = return_json

    def markets(self):
        url = self.url_for('market')
        data = self.get(url)
        return data['data']

    def ticker(self, market_id: str):
        url = self.url_for('ticker')
        params = {
            'market': str(market_id)
        }
        data = self.get(url, params=params)
        if self.return_json:
            return data
        return _m.Ticker.create_from_json(data['data'])

    def order_book(self,
                   market_id: str,
                   order_type: str,
                   page: int=None,
                   limit: int=_c.ORDERS_LIMIT):
        params = {
            'market': str(market_id),
            'type': str(order_type),
            'page': page,
            'limit': limit
        }
        url = self.url_for('book')
        data = self.get(url, params=params)
        if self.return_json:
            return data
        return _m.OrderBook.create_from_json(
            data['data'], data['pagination'])

    def trades(self,
               market_id: str,
               start: datetime=None,
               end: datetime=None,
               page: int=None,
               limit: int=_c.ORDERS_LIMIT):
        if isinstance(start, datetime):
            start = start.strftime('%Y-%m-%d')
        if isinstance(end, datetime):
            end = end.strftime('%Y-%m-%d')
        params = {
            'market': str(market_id),
            'start': start,
            'end': end,
            'page': page,
            'limit': limit,
        }
        url = self.url_for('trades')
        data = self.get(url, params=params)
        if self.return_json:
            return data
        return _m.Trades.create_from_json(
            data['data'], data.get('pagination'))



import base64
import hashlib
import hmac
import json

# local
from .client_public_v1 import BitfinexPublic
from ..common import check_keys, clean_parameters, gen_nonce


class BitfinexAuth(BitfinexPublic):

    def __init__(self, key: str=False, secret: str=False, timeout: int=30):
        super().__init__(timeout)
        check_keys(key, secret)
        self.KEY = str(key)
        self.SECRET = str(secret)

    # INFO --------------------------------------------------------------------
    # Return information about your account (trading fees).
    def account_info(self):
        return self._sign_and_get('account_infos')

    # Return information about your account (trading fees).
    def summary(self):
        return self._sign_and_get('summary')

    # Check the permissions of the key being used to generate this request.
    def key_info(self):
        return self._sign_and_get('key_info')

    # See your trading wallet information for margin trading.
    def margin_info(self):
        return self._sign_and_get('margin_infos')

    # See your balances.
    def balances(self):
        return self._sign_and_get('balances')

    # MOVEMENTS ---------------------------------------------------------------
    # Return your deposit address to make a new deposit.
    def new_deposit(self,
                    method: str,
                    wallet_name: str,
                    renew: bool=None):
        payload = {
            'method': method,
            'wallet_name': wallet_name,
            'renew': renew if renew is not None else None,
        }
        return self._sign_and_post('deposit/new', payload)

    # Allow you to move available balances between your wallets.
    def transfer(self,
                 amount: float,
                 currency: str,
                 wallet_from: str,
                 wallet_to: str):
        payload = {
            'amount': str(amount),
            'currency': str(currency),
            'walletfrom': wallet_from,
            'walletto': wallet_to,
        }
        return self._sign_and_post('transfer', payload)

    # Allow you to request a withdrawal from one of your wallet.
    def withdraw(self,
                 w_type: str,
                 wallet: str,
                 amount: float,
                 address: str):
        payload = {
            'withdraw_type': str(w_type),
            'walletselected': wallet,
            'amount': str(amount),
            'address': address,
        }
        return self._sign_and_post('withdraw', payload)

    # ORDERS ------------------------------------------------------------------
    # Submit a new order.
    def place_order(self,
                    amount: float,
                    price: float,
                    side: str,
                    ord_type: str,
                    symbol: str,
                    params: dict=None):
        payload = {
            'symbol': str(symbol),
            'amount': str(amount),
            'price': str(price),
            'side': str(side),
            'type': str(ord_type),
            'ocoorder': False,
        }
        payload.update(params or {})
        return self._sign_and_post('order/new', payload)

    # Submit a new order.
    def place_oco_order(self,
                        amount: float,
                        price: float,
                        side: str,
                        ord_type: str,
                        symbol: str,
                        buy_price_oco: float,
                        sell_price_oco: float):
        oco = {
            'ocoorder': True,
            'buy_price_oco': str(buy_price_oco),
            'sell_price_oco': str(sell_price_oco),
        }
        return self.place_order(amount, price, side, ord_type, symbol, oco)

    # Cancel an order.
    def delete_order(self, order_id: int):
        payload = {
            'order_id': order_id,
        }
        return self._sign_and_post('order/cancel', payload)

    # Cancel all orders.
    def delete_all_order(self):
        return self._sign_and_post('order/cancel/all')

    # Get the status of an order. Is it active? Was it cancelled?
    # To what extent has it been executed? etc.
    def status_order(self, order_id: int):
        payload = {
            'order_id': order_id,
        }
        return self._sign_and_post('order/status', payload)

    # View your active orders.
    def active_orders(self):
        return self._sign_and_post('orders')

    # View your latest inactive orders.
    # Limited to last 3 days and 1 request per minute.
    def orders_history(self, limit: int):
        payload = {
            'limit': limit,
        }
        return self._sign_and_post('order/hist', payload)

    # POSITIONS ---------------------------------------------------------------
    # View your active positions.
    def active_positions(self):
        return self._sign_and_post('positions')

    # Claim a position.
    def claim_position(self, position_id: int):
        payload = {
            'position_id': position_id,
        }
        return self._sign_and_post('position/claim', payload)

    # HISTORICAL DATA ---------------------------------------------------------
    # View all of your balance ledger entries.
    def balance_history(self,
                        currency: str,
                        since: float=None,
                        until: float=None,
                        limit: int=None,
                        wallet: str=None):
        payload = {
            'currency': str(currency),
            'since': since,
            'until': until,
            'limit': limit,
            'wallet': wallet,
        }
        return self._sign_and_post('history', payload)

    # View your past deposits/withdrawals.
    def movements(self,
                  currency: str,
                  method: str=None,
                  since: float=None,
                  until: float=None,
                  limit: int=None):
        payload = {
            'currency': str(currency),
            'method': method,
            'since': since,
            'until': until,
            'limit': limit,
        }
        return self._sign_and_post('history/movements', payload)

    # View your past trades.
    def past_trades(self,
                    symbol: str,
                    timestamp: float=None,
                    until: float=None,
                    limit_trades: int=None,
                    reverse: bool=None):
        payload = {
            'symbol': str(symbol),
            'timestamp': timestamp,
            'until': until,
            'limit_trades': limit_trades,
            'reverse': reverse if reverse is not None else None,
        }
        return self._sign_and_post('mytrades', payload)

    # MARGIN FUNDING ----------------------------------------------------------
    # Submit a new Offer.
    def place_offer(self,
                    currency: str,
                    amount: float,
                    rate: float,
                    period: int,
                    direction: str):
        payload = {
            'currency': str(currency),
            'amount': str(amount),
            'rate': str(rate),
            'period': period,
            'direction': str(direction),
        }
        return self._sign_and_post('offer/new', payload)

    # Cancel an offer.
    def cancel_offer(self, offer_id: int):
        payload = {
            'offer_id': offer_id,
        }
        return self._sign_and_post('offer/cancel', payload)

    # Get the status of an offer. Is it active? Was it cancelled?
    # To what extent has it been executed? etc.
    def status_offer(self, offer_id: int):
        payload = {
            'offer_id': offer_id,
        }
        return self._sign_and_post('offer/status', payload)

    # View your active offers.
    def active_offers(self):
        return self._sign_and_post('offers')

    # PRIVATE METHODS ---------------------------------------------------------
    # Pack and sign the payload of the request.
    def _sign_payload(self, path: str, payload: dict):

        payload['request'] = path
        payload['nonce'] = gen_nonce()
        payload = clean_parameters(payload)

        j = json.dumps(payload).encode('utf-8')
        encoded_body = base64.standard_b64encode(j)

        h = hmac.new(self.SECRET.encode('utf-8'), encoded_body, hashlib.sha384)
        signature = h.hexdigest()

        return {
            'X-BFX-APIKEY': self.KEY,
            'X-BFX-SIGNATURE': signature,
            'X-BFX-PAYLOAD': encoded_body,
        }

    # Packs and sign the payload and send the request with GET.
    def _sign_and_get(self, path: str, payload: dict=None):
        url, path = self.url_path_for(path)
        signed_payload = self._sign_payload(path, payload or {})
        return self.get(url, headers=signed_payload)

    # Packs and sign the payload and send the request with POST.
    def _sign_and_post(self, path: str, payload: dict=None):
        url, path = self.url_path_for(path)
        signed_payload = self._sign_payload(path, payload or {})
        return self.post(url, headers=signed_payload, data=payload)
